# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 17:03:41 2021

@author: richa
"""
import functools
import tkinter as tk

from desktop_shop import gui, server
from desktop_shop.gui import callbacks, components


# pylint: disable=no-member
class View(components.View):
    """Checkout view"""

    @classmethod
    def create(cls, *_):
        """Initialises an empty checkout view.

        Currently, all checkout view contents are dynamically generated in the function
        init_checkout_data_in_checkout_view
        """
        return cls()

    def init_checkout(self):
        """Dynamically generates all contents of the checkout view"""
        chosen_product_ids = gui.app.data["cart"]
        gui.app.builder.view = self

        for frame in self._frame_components.values():
            frame.hide()
        self._frame_components = {}

        with gui.db_conn as cursor:
            args = cursor, chosen_product_ids
            product_datas = server.query_product_data_from_product_table_by_product_ids(
                *args
            )

        data_packets = [
            (str(id_), name, price)
            for id_, name, price in product_datas
            if str(id_) in chosen_product_ids
        ]

        i = 0
        for i, (product_id, name, price) in enumerate(data_packets, 1):
            gui.app.builder.root = gui.root
            frame = gui.app.builder.create(
                "frame", name=product_id, relief=tk.RAISED, bd=3
            )
            frame.place(row=i, row_span=1, col=1, col_span=3, sticky="we")
            gui.app.builder.root = frame.component.tk_component

            # product information label
            full_product_name = f"{name} (#{product_id}):"
            full_text = f"#{i:<3} {full_product_name:<25} ${price:<7}"
            gui.app.builder.create("label", width=42, text=full_text).place(
                row=i, col=0, sticky="w"
            )

            # remove from cart button
            callback = functools.partial(self.remove_from_cart, product_id)
            button = gui.app.builder.create(
                "button", text="Remove from cart", command=callback, bd=3
            )
            button.place(row=i, col=2, sticky="e")

        # confirm transaction button
        gui.app.builder.root = gui.root
        frame = gui.app.builder.create("frame")
        frame.place(row=i + 1, col=3)
        gui.app.builder.root = frame.component.tk_component
        gui.app.builder.create(
            "button2",
            name="checkout",
            text="Confirm transaction",
            command=self.place_order,
            bd=3,
        ).place(sticky="we")

    def remove_from_cart(self, product_id):
        """Callback for dynamically generated remove product from cart button. Gets
        the product to remove directly from the button_name (kind of hacky...),
        removes all dynamically generated widgets on the same row as the button and then
        repacks all other components to fill the resulting gap
        """
        gui.app.data["cart"].remove(product_id)
        frame = self._frame_components.pop(f"{product_id}_frame")
        frame.hide()

        for i, frame in enumerate(self._frame_components.values(), 1):
            frame.row = i
        gui.app["checkout"].repack()

        if not gui.app.data["cart"]:
            gui.app["main_menu"].hide_components("checkout_button")
            callbacks.switch_to_home()

    @staticmethod
    def place_order():
        """Gets all product ids stored in the user cart and sends a transaction request to the
        server. If the server does not answer with a valid session id, the transaction failed
        and an error message is shown, else a confirmation message is shown to the user
        """
        user_email = gui.app.data["user_data"].email
        chosen_product_ids = gui.app.data["cart"]
        if user_email and chosen_product_ids:
            session_id = gui.app.data["session_id"]

            with gui.db_conn as cursor:
                # pylint: disable=redundant-keyword-arg
                # pylint: disable=unexpected-keyword-arg
                # pylint: disable=unpacking-non-sequence
                new_session_id, _ = server.add_transaction(
                    cursor,
                    user_email,
                    chosen_product_ids,
                    user_email=user_email,
                    session_id=session_id,
                )

            gui.app.data["session_id"] = new_session_id
            if new_session_id is not None:
                callbacks.show_message("We have placed your order.")
                gui.app.data["cart"] = []
                gui.app.views_dict["checkout"].clear()
                gui.app["main_menu"].hide_components("checkout_button")
            else:
                callbacks.show_error_message("Your session has expired.")
        else:
            callbacks.show_error_message("We were not able to place your order.")
        callbacks.switch_to_home()



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 25 20:07:42 2020

@author: Korean_Crimson
"""
import hmac

from . import crypto, database, util

# combined with every salt for extra security in pw hashing
PEPPER = "secret"


class ServerSideLogInDataError(Exception):
    """Exception for login failures"""

    def __init__(self):
        super().__init__(
            "Login data invalid. Corrupted login data was retrieved from the server."
        )


def verify_database_call(function):
    """Function decorator for database requests that work on sensitive data.
    This decorator mandates sending all the normal arguments, as well as two
    special keyword arguments (your linter may not like this), session_id and
    user_email. If the specified session_id cannot be found in the sessions table
    for the user identified by the user_email, the database request is denied.
    """

    def wrapper(*args, user_email="", session_id=""):
        cursor, *_ = args
        verified = database.verify_session_id_by_user_email(
            cursor, session_id, user_email
        )
        if verified:
            results = function(*args)

            # get new session id and add it to database
            new_session_id = _add_new_session(cursor, user_email)
        else:
            new_session_id = None
            results = None
        return new_session_id, results

    return wrapper


@verify_database_call
def query_user_data(cursor, user_email):
    """Queries the user data for the specified user_email from the users table.

    As this is a verified database call, a valid session id needs to be passed
    in (see decorator def) in order for the database request to succeed.
    """
    return database.query_user_data_by_user_email(cursor, user_email)


@verify_database_call
def add_transaction(cursor, user_email, chosen_product_ids):
    """Adds a new transaction containing the passed product ids for specified user.

    As this is a verified database call, a valid session id needs to be passed
    in (see decorator def) in order for the database request to succeed.
    """
    user_id = database.query_user_id_from_user_email(cursor, user_email)
    date = util.get_current_date()
    transaction_data = [user_id, date]
    return database.add_transaction(cursor, transaction_data, chosen_product_ids)


@verify_database_call
def update_user(cursor, user_data, user_email):
    """Updates the user data with the passed data

    As this is a verified database call, a valid session id needs to be passed
    in (see decorator def) in order for the database request to succeed.
    """
    return database.update_user_by_user_email(cursor, list(user_data), user_email)


@verify_database_call
def update_user_password(cursor, password, user_email):
    """Updates the user password with the passed data.

    As this is a verified database call, a valid session id needs to be passed
    in (see decorator def) in order for the database request to succeed.
    """
    return database.update_user_password(cursor, password, user_email, PEPPER)


def query_product_data_from_product_table(cursor):
    """Queries all product data from products table"""
    return database.query_product_data_from_product_table(cursor)


def query_product_data_from_product_table_by_product_ids(cursor, product_ids):
    """Queries product data from the products table by ids"""
    return database.query_product_data_from_product_table_by_product_ids(
        cursor, product_ids
    )


def add_user(cursor, user_data, password):
    """Adds a new user to the users table, with the user_data specified"""
    database.add_user(cursor, list(user_data), password, PEPPER)
    new_session_id = _add_new_session(cursor, user_data.email)
    return new_session_id


def login(cursor, user_email, password):
    """Gets the pw salt and pw hash from the database for the specified user.
    If they match the user_email and password passed into the login function,
    a new session id is generated, stored in the sessions table and returned.
    """
    data = database.query_pw_hash_and_salt_by_user_email(cursor, user_email)
    if len(data) == 3:
        # pylint: disable=unbalanced-tuple-unpacking
        pw_salt, pw_hash, hash_function_name = data
        hash_function = crypto.get_hash_function_from_string(hash_function_name)
        if hash_function is None:
            raise ServerSideLogInDataError

        hashed_password = hash_function.hash(password, pw_salt + PEPPER)
        if hmac.compare_digest(hashed_password, pw_hash):
            new_session_id = _add_new_session(cursor, user_email)
            return new_session_id
    return None


def _add_new_session(cursor, user_email):
    new_session_id = crypto.generate_new_session_id()
    timestamp = util.generate_timestamp()
    user_id = database.query_user_id_from_user_email(cursor, user_email)
    session_data = [new_session_id, user_id, timestamp]
    database.add_session(cursor, session_data)
    return new_session_id



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 16:31:46 2020

@author: Korean_Crimson
"""

import datetime


def generate_timestamp() -> str:
    """Generates date+timestamp of current time"""
    return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")


def get_current_date() -> str:
    """Returns current date"""
    return datetime.datetime.now().strftime("%Y-%m-%d")


def validate_date_string(date_string) -> bool:
    """Checks if passed string can be converted to date"""
    try:
        datetime.datetime.strptime(date_string, "%Y-%m-%d")
        valid = True
    except ValueError:
        valid = False
    return valid



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 09:52:44 2021

@author: Korean_Crimson
"""
import os

from desktop_shop.datagen import generate_data

from .gui import app, db_conn, init

# fast db generation if not present
if not os.path.isdir("main.db"):
    generate_data.generate(
        hash_iterations=1, transactions=100_000, users=10_000, products=20
    )

try:
    init.init()
    app.views_dict["main_menu"].pack()
    app.views_dict["home"].pack()
    app.mainloop()
finally:
    db_conn.close()



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 09:53:07 2021

@author: Korean_Crimson
"""
import re

from desktop_shop import gui, server, user, util

# pylint: disable=E1123,E1124


def login(password=None, email=None):
    """Tries to get a fresh session id from the server by sending password and
    user_email to it. If a new session is granted (password and user_email match),
    gui switches to logged-in home view (login/register button hide, logged-in-as shows)
    """
    password = gui.app["login"]["pw_entry"].get() if password is None else password
    user_email = gui.app["login"]["email_entry"].get() if email is None else email
    with gui.db_conn as cursor:
        session_id = server.login(cursor, user_email, password)

    if session_id is None:
        gui.app.views_dict["login"].unhide_components("login_failed_label")
        return

    gui.app.data["session_id"] = session_id
    gui.app["login"].hide_components("login_failed_label")
    switch_to_home()

    # pylint: disable=unbalanced-tuple-unpacking
    with gui.db_conn as cursor:
        session_id, user_data = server.query_user_data(
            cursor, user_email, user_email=user_email, session_id=session_id
        )

    gui.app.data["session_id"] = session_id
    store_user_data(user_data)
    gui.app["main_menu"].unhide_components("logged_in_as_frame")
    gui.app["main_menu"].hide_components(
        "message_frame", "login_button", "register_button"
    )
    set_logged_in_as_user_text()
    gui.app["login"].clear_entries()


def sign_out():
    """Signs out the currently logged in user. clears the user data, login data,
    register data, checkout basket and returns to not-logged-in home view
    (unhides register/login buttons, hides logged-in-as message)
    """
    gui.app.data = {
        "session_id": None,
        "user_data": user.UserSignUpData(),
        "pw_hash": "",
        "cart": [],
    }
    gui.app["login"].clear_entries()
    gui.app["register"].clear_entries()
    gui.app["checkout"].clear()
    gui.app["main_menu"].hide_components(
        "logged_in_as_frame", "message_frame", "checkout_button"
    )
    gui.app["main_menu"].unhide_components("login_button", "register_button")
    switch_to_home()


def set_logged_in_as_user_text():
    """Sets text of options_dropdown in main menu view to --> Logged in as {full_name}"""
    full_name = gui.app.data["user_data"].full_name
    gui.app["main_menu"]["options_dropdown"].set_var(f"Logged in as {full_name}")


def register():
    """Gets all data entered in the register view and validates it. If valid,
    send the data and the current date to the server and add it to a new user
    in the database, else show an error message
    """
    user_data = gui.app["register"].get_user_data()
    valid_data = validate_user_data(user_data)

    password = gui.app["register"]["pw_entry"].get_var()
    confirm_password = gui.app["register"]["confirm_pw_entry"].get_var()
    valid_password = validate_password(password, confirm_password)

    if not valid_data or not valid_password:
        return

    with gui.db_conn as cursor:
        session_id = server.add_user(cursor, user_data, password)

    gui.app.data["session_id"] = session_id
    if session_id is not None:
        login(password, user_data.email)
        gui.app["register"].clear_entries()
    else:
        show_error_message("Failed to register.")


def edit_user_data():
    """Edits the user data (callback for edit in profile view). Gets all user
    data from the text entries and validates it using validate_user_data. If
    valid, the data is sent to the server. If the server does not respond with
    a valid new session id, the changes failed and a corresponding error message
    is shown.
    """
    user_data = gui.app["profile"].get_user_data()
    valid_data = validate_user_data(user_data)
    if not valid_data:
        return

    session_id = gui.app.data["session_id"]

    with gui.db_conn as cursor:
        user_data_ = user_data[:-1]  # ignore join date
        # pylint: disable=unpacking-non-sequence
        new_session_id, *_ = server.update_user(
            cursor,
            user_data_,
            user_data.email,
            user_email=user_data.email,
            session_id=session_id,
        )

    gui.app.data["session_id"] = new_session_id
    gui.app.data["user_data"] = user_data
    populate_profile_with_user_data()
    if not new_session_id:
        show_error_message("Failed to edit data.")


def edit_user_password():
    """Edits the user password (callback for edit in profile view). Gets the
    password and confirm passwords and validates them. If they are valid, the
    new data is sent to the serer. If the server does not respond with a new
    valid session id, the password change failed and the user is shown an error
    message, else a confirmation message is shown.
    """
    password = gui.app["profile"]["pw_entry"].get_var()
    confirm_password = gui.app["profile"]["confirm_pw_entry"].get_var()
    valid_password = validate_password(password, confirm_password)
    if not valid_password:
        return

    session_id = gui.app.data["session_id"]
    user_email = gui.app.data["user_data"].email

    with gui.db_conn as cursor:
        # pylint: disable=unpacking-non-sequence
        new_session_id, *_ = server.update_user_password(
            cursor, password, user_email, user_email=user_email, session_id=session_id
        )

    gui.app.data["session_id"] = new_session_id
    if not new_session_id:
        show_error_message("Failed to edit password.")
        return

    show_message("Set new password successfully.")
    gui.app["profile"]["confirm_pw_entry"].set_var("")
    gui.app["profile"]["pw_entry"].set_var("")
    gui.app["profile"].hide_components("password_change_frame")
    gui.app["profile"].unhide_components("password_change_frame_button")


def validate_user_data(user_data):
    """Validates all the data entered in the register view. Validates that the
    email is in the correct format, validates that the first and last name are
    alphabetic and not empty, validates that the gender is either m or f and
    validates that the date of birth is a valid date and is not empty
    """
    # validate email
    found_email = re.findall(".+?@.+\\..+", user_data.email)
    if not found_email:
        show_error_message("Email needs to be of the format address@domain.")
        return False

    # validate first_name
    found_first_name = re.findall("\\w+?", user_data.first_name)
    if not found_first_name:
        show_error_message("First name needs to be alphabetic.")
        return False

    # validate last_name
    found_last_name = re.findall("\\w+?", user_data.last_name)
    if not found_last_name:
        show_error_message("Last name needs to be alphabetic.")
        return False

    # validate gender
    found_gender = re.findall("[mf]", user_data.gender)
    if user_data.gender and not found_gender:
        show_error_message("Gender needs to be m or f.")
        return False

    # validate dob
    if user_data.dob and not util.validate_date_string(user_data.dob):
        show_error_message("Date of birth needs to be in the format YYYY-MM-DD.")
        return False

    return True


def validate_password(password, confirm_password):
    """Validates that the passwords match and are longer than 8 characters (returns bool)"""
    if len(password) < 8:
        show_error_message("Password needs to be at least 8 characters long.")
        return False

    if not password == confirm_password:
        show_error_message("Passwords don't match.")
        return False

    return True


def show_message(message):
    """Shows the specified message in the main menu view"""
    gui.app["main_menu"]["message_label"].set_var(message)
    gui.app["main_menu"].hide_components("error_message_label")
    gui.app["main_menu"].unhide_components("message_label", "message_frame")


def show_error_message(message):
    """Shows the specified error message in the main menu view"""
    gui.app["main_menu"]["error_message_label"].set_var(message)
    gui.app["main_menu"].hide_components("message_frame")
    gui.app["main_menu"].unhide_components("error_message_label", "message_frame")


def show_password_change_frame():
    """Callback for change password button in profile view, unhides the password change frame"""
    gui.app["profile"].unhide_components("password_change_frame")
    gui.app["profile"].hide_components("password_change_frame_button")


def store_user_data(user_data):
    """Stores the passed user data in the appropriate gui data fields"""
    if len(user_data) == 6:
        gui.app.data["user_data"] = user.UserSignUpData(*user_data)
    else:
        sign_out()
        show_error_message("Something went wrong while logging in. 2")


def switch_to_home():
    """Switches to home view"""
    gui.app.switch_to("home", "main_menu")


def switch_to_login():
    """Switches to login view"""
    gui.app.switch_to("login", "main_menu")


def switch_to_register():
    """Switches to register view"""
    gui.app.switch_to("register", "main_menu")


def switch_to_checkout():
    """Switches to checkout view. If no valid session id is stored, a warning is given to user"""
    gui.app["checkout"].init_checkout()
    gui.app.switch_to("checkout", "main_menu")
    if gui.app.data["session_id"] is None:
        show_error_message("Please login to place your order")


def switch_to_profile():
    """Switches to the profile view after populating it with the currently logged in user data"""
    populate_profile_with_user_data()
    gui.app.switch_to("profile", "main_menu")


def activate_profile_entry(entry_name):
    """Callback for edit button for the respective field in the user data profile view"""
    gui.app["profile"][entry_name].config(state="normal")
    gui.app["profile"]["edit_user_data_button"].config(state="normal")


def on_dropdown_value_write_event(*_):
    """Callback for options dropdown selection, either switches to profile view or signs out"""
    selected_option = gui.app["main_menu"]["options_dropdown"].get_var()
    if selected_option == "Profile":
        switch_to_profile()
        set_logged_in_as_user_text()
    elif selected_option == "Sign out":
        sign_out()
    gui.app["main_menu"]["options_dropdown"].config(state="active")


def populate_profile_with_user_data():
    """Stores all user data of the currently logged in user in the profile view"""
    gui.app["profile"].store_user_data(gui.app.data["user_data"])
    widget_names = [
        "first_name_entry",
        "last_name_entry",
        "gender_entry",
        "date_joined_data_label",
        "email_entry",
        "dob_entry",
        "edit_user_data_button",
    ]
    for widget_name in widget_names:
        gui.app["profile"][widget_name].config(state="disabled")



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 09:52:58 2021

@author: Korean_Crimson
"""
import sqlite3

from desktop_shop.gui.components import Gui, Tk

root = Tk()
app = Gui(root)
db_conn = sqlite3.connect("main.db")



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 16 20:19:46 2020

@author: Korean_Crimson
"""
import enum
import hashlib
import random
import secrets
from dataclasses import dataclass
from typing import Callable, Dict, Optional

HashFunctionCallable = Callable[[str, str, int], str]
BASE62 = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


class InvalidHashFunctionException(ValueError):
    """Exception thrown when an invalid hash function is specified"""


class HashFunctions(enum.Enum):
    """hash functions enum (values are functions)"""

    HASHLIB_PBKDF2_HMAC_SHA256 = enum.auto()


@dataclass
class HashFunction:
    """Contains a hash function enum and its iterations keyword.
    Encapsulates a hash operation.
    """

    function: HashFunctions
    iterations: int

    def hash(self, password: str, salt: str) -> str:
        """Returns the result of the hash function hashing the provided password and salt"""
        function = get_function_from_hash_function_enum(self.function)
        return function(password, salt, self.iterations)

    def __str__(self):
        return f"{self.function.name}::{self.iterations}"


def hash_string(password: str, salt: str, iterations: int = 100_000) -> str:
    """Hashes the specified password with the passed salt.
    Returns hash in hex format (str).
    """
    encoded_password = bytes(password, encoding="utf-8")
    encoded_salt = bytes(salt, encoding="utf-8")
    derived_key = hashlib.pbkdf2_hmac(
        "sha256", encoded_password, encoded_salt, iterations=iterations
    )
    return derived_key.hex()


def get_hash_function(iterations: int) -> HashFunction:
    """Returns a HashFunction object"""
    return HashFunction(
        function=HashFunctions.HASHLIB_PBKDF2_HMAC_SHA256, iterations=iterations
    )


def get_hash_function_from_string(hash_function_name: str) -> Optional[HashFunction]:
    """Looks up a hash function in the HashFunctions enum from a str key in
    the format name::iterations.
    """
    name, iterations = hash_function_name.split("::")
    hash_function = HashFunctions.__dict__.get(name)
    if hash_function is not None:
        return HashFunction(function=hash_function, iterations=int(iterations))
    return None


def get_function_from_hash_function_enum(
    hash_function: HashFunctions,
) -> HashFunctionCallable:
    """Looks up the passed HashFunctions enum and returns a corresponding function handle"""
    functions: Dict[HashFunctions, HashFunctionCallable] = {
        HashFunctions.HASHLIB_PBKDF2_HMAC_SHA256: hash_string
    }
    hash_function_callable = functions.get(hash_function)
    if hash_function_callable is None:
        raise InvalidHashFunctionException("Hash function could not be found!")
    return hash_function_callable


def generate_new_salt() -> str:
    """Generates a new base 62 encoded salt based on a cryptographically
    secure random number, to hash strings.
    """
    rng = random.SystemRandom()
    random_number = rng.randint(0, 2 ** 64)
    salt = _b62encode(random_number)
    return salt


def _b62encode(num: int, alphabet: str = BASE62) -> str:
    """Encode a positive number into Base 62 and return the string."""
    char_array = [] if num else [alphabet[0]]
    base = len(alphabet)
    while num:
        num, rem = divmod(num, base)
        char_array.append(alphabet[rem])
    char_array.reverse()
    encoded_string = "".join(char_array)
    return encoded_string


def generate_new_session_id() -> str:
    """Creates a cryptographically-secure, URL-safe string"""
    return secrets.token_urlsafe(16)



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 17:29:20 2020

@author: Korean_Crimson
"""

import datetime
import random
from typing import Dict, List, Tuple

import requests  # type: ignore
import wikipedia
from bs4 import BeautifulSoup

DateTuple = Tuple[int, int, int]


def fetch_first_names() -> Dict[str, str]:
    """fetches first names from the web and returns them as a dict in the
    format first_name: gender
    """
    page = requests.get("https://www.ssa.gov/oact/babynames/decades/century.html")
    soup = BeautifulSoup(page.content, "html.parser")
    first_names = {}
    for row in soup.find("table").select("tr")[2:-1]:
        data = [data.get_text() for data in row.select("td")]
        _, male, _, female, _ = data
        first_names[male] = "m"
        first_names[female] = "f"
    return first_names


def fetch_last_names() -> List[str]:
    """fetches last names from the web and returns them in a list"""
    wiki = wikipedia.page("List_of_most_common_surnames_in_Europe")
    soup = BeautifulSoup(wiki.html(), "html.parser")

    last_names = set()
    for table in soup.find_all("table"):
        for row in table.select("tr"):
            data = [data.get_text() for data in row.select("td")]
            if len(data) == 3:
                _, last_name, _ = data
                if last_name.isalpha():
                    last_names.add(last_name)
    return list(last_names)


def get_random_date(start: DateTuple, end: DateTuple) -> str:
    """start and end each need to be a list/tuple of length 3 (Y, M, D)"""
    start_date = datetime.date(*start)
    end_date = datetime.date(*end)
    time_between_dates = end_date - start_date
    random_number_of_days = random.randrange(time_between_dates.days)
    time_diff = datetime.timedelta(days=random_number_of_days)
    random_date = start_date + time_diff
    return random_date.strftime("%Y-%m-%d")



# -*- coding: utf-8 -*-
"""
Created on Sun Nov 22 14:31:48 2020

@author: Korean_Crimson
"""
from desktop_shop import crypto


def query_user_id_from_user_email(cursor, user_email):
    """Queries for the user id of the user specified by the user_email passed (unique)"""
    command = "SELECT user_id FROM users WHERE email_address = ?"
    user_ids = [user_id for user_id, *_ in cursor.execute(command, [user_email])]
    return user_ids[0] if user_ids else None


def query_user_ids_from_user_table(cursor):
    """Returns all user_ids found in users table as a list"""
    user_ids = [user_id for user_id, *_ in cursor.execute("SELECT user_id FROM users")]
    return user_ids


def query_product_ids_from_product_table(cursor):
    """Returns all product ids found in products table as a list"""
    product_ids = [
        prod_id for prod_id, *_ in cursor.execute("SELECT product_id FROM products")
    ]
    return product_ids


def query_product_data_from_product_table(cursor):
    """Queries all produt data from the products table"""
    data = [list(row) for row in cursor.execute("SELECT * FROM products")]
    return data


def query_product_data_from_product_table_by_product_ids(cursor, product_ids):
    """Queries product data from the products table, by the passed product ids"""

    # sanitize input and make sure all product ids are only numeric
    if not all(product_id.isnumeric() for product_id in product_ids):
        return []

    joined_product_ids = ",".join([str(product_id) for product_id in product_ids])
    joined_product_ids_str = f"({joined_product_ids})"

    # This .format looks like an SQL injection vulnerability, but there is currently
    # no other way to execute Sqlite "WHERE something IN list" statements in Python.
    # As we make sure the passed list only contains numeric strings, this should be OK.
    command = "SELECT * FROM products WHERE product_id IN " + joined_product_ids_str
    data = [list(row) for row in cursor.execute(command)]
    return data


def query_product_price_from_product_table(cursor, product_ids):
    """Returns all prices found for the specified product ids as a list"""

    # sanitize input and make sure all product ids are only numeric
    if not all(product_id.isnumeric() for product_id in product_ids):
        return []

    joined_product_ids = ",".join([str(product_id) for product_id in product_ids])
    joined_product_ids_str = f"({joined_product_ids})"

    command = "SELECT price FROM products WHERE product_id IN " + joined_product_ids_str
    product_prices = [product_price for product_price, *_ in cursor.execute(command)]
    return product_prices


def query_user_data_by_user_email(cursor, user_email):
    """Queries the user data from the users table, by the user identified by the user_email"""
    command = """SELECT
                    first_name,
                    last_name,
                    gender,
                    dob,
                    email_address,
                    join_date
                FROM users
                WHERE email_address = ?"""

    data = cursor.execute(command, [user_email])
    data = list(data)
    return list(data[0]) if data else []


def query_user_data(cursor, user_id):
    """Queries the data for the user from the users table, by the user id passed"""
    command = """'SELECT
                    first_name,
                    last_name,
                    gender,
                    dob,
                    email_address,
                    join_date,
                FROM users
                WHERE user_id = ?"""

    data = cursor.execute(command, [str(user_id)])
    data = list(data)
    return list(data[0]) if data else []


def query_pw_hash_and_salt_by_user_email(cursor, user_email):
    """Queries the password hash, salt and hashing function from the users table
    for the specified user_email
    """
    command = (
        "SELECT pw_salt, pw_hash, hash_function FROM users WHERE email_address = ?"
    )
    data = cursor.execute(command, [user_email])
    data = list(data)
    return list(data[0]) if data else []


def get_last_added_transaction_id_from_transactions_table(cursor):
    """Returns the id of the transaction last added to the transactions table"""
    transaction_id, *_ = [
        id_ for id_, *_ in cursor.execute("""SELECT last_insert_rowid()""")
    ]
    return transaction_id


def verify_session_id(cursor, session_id, user_id):
    """Verifies that the passed session_id is held by a user identified by the passed user id"""
    command = """SELECT session_id
                FROM sessions
                WHERE session_id = ?
                AND user_id = ?"""

    data = cursor.execute(command, [session_id, user_id])
    data = [d for d, *_ in data]
    verified = data and data[0] == session_id

    # DONT REMOVE: for some reason, without this useless if, the function doesnt work....
    if not verified:
        verified = False
    return verified


def verify_session_id_by_user_email(cursor, session_id, user_email):
    """Verifies that the passed session_id is held by a user identified by the passed user_email"""
    command = """SELECT session_id FROM sessions
                  WHERE session_id = ?
                  AND user_id IN
                  (
                      SELECT user_id FROM users
                      WHERE email_address = ?)"""

    data = cursor.execute(command, [session_id, user_email])
    data = [d for d, *_ in data]
    verified = data and data[0] == session_id
    if not verified:
        verified = False
    return verified


def add_user(cursor, user_data, password, pepper="", iterations=100_000):
    """Adds a user with the specified user data to the users table"""
    command = """INSERT INTO users
                (first_name, last_name, gender, dob, email_address, join_date, pw_salt, pw_hash, hash_function)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)"""

    # hash password
    salt = crypto.generate_new_salt()
    hash_function = crypto.get_hash_function(iterations)
    hashed_password = hash_function.hash(password, salt + pepper)
    user_data = list(user_data) + [salt, hashed_password, str(hash_function)]

    cursor.execute(command, user_data)


def update_user(cursor, user_data, user_id):
    """user_data needs to be a full set of user_data, without the user_id,
    which is passed separately
    """
    user_data.append(user_id)
    command = """UPDATE users
                SET
                    first_name = ?,
                    last_name = ?,
                    gender = ?,
                    dob = ?,
                    email_address = ?
                WHERE user_id = ?"""
    cursor.execute(command, user_data)


def update_user_password(cursor, password, user_email, pepper="", iterations=100_000):
    """Updates the password hash in the users table for the user specified"""
    salt = crypto.generate_new_salt()
    pw_hash = crypto.get_hash_function(iterations).hash(password, salt + pepper)
    command = "UPDATE users SET pw_hash = ?, pw_salt = ? WHERE email_address = ?"
    cursor.execute(command, [pw_hash, salt, user_email])


def update_user_by_user_email(cursor, user_data, user_email):
    """user_data needs to be a full set of user_data, without the user_id,
    which is passed separately
    """
    user_data.append(user_email)
    command = """UPDATE users
                SET
                    first_name = ?,
                    last_name = ?,
                    gender = ?,
                    dob = ?,
                    email_address = ?
                WHERE email_address = ?"""
    cursor.execute(command, user_data)


def add_transaction(cursor, transaction_data, chosen_product_ids):
    """adds a transaction with the specified transaction data to the transactions
    table and the detailed transactions table.
    """
    product_prices = query_product_price_from_product_table(cursor, chosen_product_ids)
    cost = sum(product_prices)
    transaction_data.append(cost)

    command = "INSERT INTO transactions (user_id, date, cost) VALUES (?, ?, ?)"
    cursor.execute(command, transaction_data)

    transaction_id = get_last_added_transaction_id_from_transactions_table(cursor)

    command = (
        "INSERT INTO detailed_transactions (transaction_id, product_id) VALUES (?, ?)"
    )
    for chosen_product_id in chosen_product_ids:
        cursor.execute(command, [transaction_id, chosen_product_id])


def add_session(cursor, session_data):
    """adds a session with the specified session data to the session table"""
    command = "INSERT INTO sessions (session_id, user_id, timestamp) VALUES (?, ?, ?)"
    cursor.execute(command, session_data)


def add_product(cursor, product_data):
    """adds a product with the specified product data to the products table"""
    command = "INSERT INTO products (name, price) VALUES (?, ?)"
    cursor.execute(command, product_data)


def create_user_table(cursor):
    """Creates user table. Only called in generate_database"""
    # remove table
    cursor.execute("""DROP TABLE IF EXISTS users""")

    # create table
    cursor.execute(
        """CREATE TABLE users (
                    user_id INTEGER PRIMARY KEY,
                    first_name TEXT NOT NULL,
                    last_name TEXT NOT NULL,
                    gender TEXT,
                    dob TEXT CHECK(CAST(strftime('%s', join_date)  AS  integer) >
                                   CAST(strftime('%s', dob)  AS  integer)),
                    email_address TEXT NOT NULL UNIQUE COLLATE NOCASE,
                    join_date TEXT NOT NULL,
                    pw_salt NOT NULL,
                    pw_hash NOT NULL,
                    hash_function NOT NULL
                    CHECK (email_address LIKE '%_@_%._%')
                    )"""
    )

    # remove index based on user id
    cursor.execute("""DROP INDEX IF EXISTS idx_user_id""")

    # create index based on user_id
    cursor.execute("""CREATE INDEX idx_user_id ON users(user_id)""")


def create_transactions_table(cursor):
    """Creates transactions table. Only called in generate_database"""
    # remove table
    cursor.execute("""DROP TABLE IF EXISTS transactions""")

    # create table
    cursor.execute(
        """CREATE TABLE transactions (
                    transaction_id INTEGER PRIMARY KEY,
                    user_id INTEGER,
                    date TEXT,
                    cost REAL NOT NULL,
                    FOREIGN KEY (user_id)
                        REFERENCES users (user_id)
                        ON UPDATE CASCADE
                        ON DELETE RESTRICT
                    )"""
    )

    # drop index based on user_id
    cursor.execute("""DROP INDEX IF EXISTS idx_user_id""")

    # create index based on user_id
    cursor.execute("""CREATE INDEX idx_user_id ON transactions(user_id)""")


def create_detailed_transactions_table(cursor):
    """Creates detailed transactions table. Only called in generate_database"""
    # remove table
    cursor.execute("""DROP TABLE IF EXISTS detailed_transactions""")

    # create table
    cursor.execute(
        """CREATE TABLE detailed_transactions (
                    transaction_id INTEGER,
                    product_id INTEGER,
                    FOREIGN KEY (transaction_id)
                        REFERENCES transactions (transaction_id)
                        ON UPDATE CASCADE
                        ON DELETE RESTRICT,
                    FOREIGN KEY (product_id)
                        REFERENCES products (product_id)
                        ON UPDATE CASCADE
                        ON DELETE RESTRICT)
                    """
    )


def create_products_table(cursor):
    """Creates products table. Only called in generate_database"""
    # remove table
    cursor.execute("""DROP TABLE IF EXISTS products""")

    # create table
    cursor.execute(
        """CREATE TABLE products (
                    product_id INTEGER PRIMARY KEY,
                    name TEXT,
                    price REAL NOT NULL
                    )"""
    )


def create_sessions_table(cursor):
    """Creates sessions table. Only called in generate_database"""
    # remove table
    cursor.execute("""DROP TABLE IF EXISTS sessions""")

    # create table
    cursor.execute(
        """CREATE TABLE sessions(
                    session_id TEXT PRIMARY KEY,
                    user_id INTEGER,
                    timestamp TEXT,
                    FOREIGN KEY (user_id)
                        REFERENCES users (user_id)
                        ON UPDATE CASCADE
                        ON DELETE CASCADE
                    )"""
    )


if __name__ == "__main__":
    import argparse

    from desktop_shop.datagen import generate_data

    parser = argparse.ArgumentParser(description="Database generation interface")
    parser.add_argument(
        "action", choices=["generate"], help="database action to be performed"
    )
    parser.add_argument(
        "--fast",
        action="store_true",
        help="reduces number of password hashing operations",
    )
    parser.add_argument(
        "--minimal", action="store_true", help="reduces size of all tables to 1"
    )
    parser.add_argument(
        "--transactions",
        nargs=1,
        default=100_000,
        type=int,
        help="pass number of transactions to be added to database",
    )
    parser.add_argument(
        "--users",
        nargs=1,
        default=10_000,
        type=int,
        help="pass number of users to be added to database",
    )
    parser.add_argument(
        "--products",
        nargs=1,
        default=20,
        type=int,
        help="pass number of products to be added to database",
    )
    args = parser.parse_args()

    if args.action == "generate":
        number_of_hashes = 1 if args.fast else 100_000  # pylint: disable=invalid-name
        transactions, users, products = (
            (1, 1, 1)
            if args.minimal
            else (args.transactions, args.users, args.products)
        )
        generate_data.generate(
            number_of_hashes, transactions=transactions, users=users, products=products
        )



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 23:08:31 2021

@author: richa
"""

from dataclasses import dataclass


# pylint: disable=function-redefined
@dataclass
class UserData:
    """Stores user data"""

    first_name: str = ""
    last_name: str = ""
    gender: str = ""
    dob: str = ""  # date of birth
    email: str = ""

    def __getitem__(self, value):
        return list(self)[value]

    def __iter__(self):
        yield self.first_name
        yield self.last_name
        yield self.gender
        yield self.dob
        yield self.email

    @property
    def full_name(self) -> str:
        """Full name getter, returns first and last names joined together"""
        return f"{self.first_name} {self.last_name}"

    @property  # type: ignore
    def first_name(self) -> str:
        """First name getter, returns capitalized first name"""
        return self._first_name.title()

    @first_name.setter
    def first_name(self, value):
        """First name setter"""
        self._first_name: str = value

    @property  # type: ignore
    def last_name(self) -> str:
        """Last name getter, returns capitalized last name"""
        return self._last_name.title()

    @last_name.setter
    def last_name(self, value):
        """Last name setter"""
        self._last_name: str = value


@dataclass
class UserSignUpData(UserData):
    """Stores user data and join date"""

    join_date: str = ""

    def __iter__(self):
        for attr in super().__iter__():
            yield attr
        yield self.join_date



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 16:47:28 2021

@author: richa
"""
from typing import Dict, Type

from desktop_shop.gui import components

from . import checkout, home, login, main_menu, profile, register


class View(components.View):
    """View Protocol interface"""

    @classmethod
    def create(
        cls, window: components.Tk, builder: components.Builder
    ) -> components.View:
        """Creates the View"""


views: Dict[str, Type[View]] = {
    "checkout": checkout.View, # type: ignore
    "home": home.View, # type: ignore
    "login": login.View, # type: ignore
    "main_menu": main_menu.View, # type: ignore
    "profile": profile.View, # type: ignore
    "register": register.View, # type: ignore
}



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 16:43:01 2021

@author: richa
"""
import tkinter as tk

from desktop_shop import user, util
from desktop_shop.gui import callbacks, components


class View(components.View):
    """Register view"""

    @classmethod
    def create(cls, window: components.Tk, builder: components.Builder):
        """Initialises register view, including all its components"""
        builder.view = cls()
        builder.root = window

        frame = builder.create("frame", name="register", relief=tk.SUNKEN, bd=2)
        frame.place(row=8, row_span=7, col=1, col_span=3, sticky="nsew")
        builder.root = frame.component.tk_component

        builder.create("label", text="Register").place(row=0, col=1)

        # email
        builder.create("label", text="*Email address").place(row=1, col=0)
        entry = builder.create("entry", name="email", textvariable=tk.StringVar())
        entry.place(row=1, col=1, sticky="nsew")

        # first name
        builder.create("label", text="*First name").place(row=2, col=0)
        entry = builder.create("entry", name="first_name", textvariable=tk.StringVar())
        entry.place(row=2, col=1, sticky="nsew")

        # last name
        builder.create("label", text="*Last name").place(row=3, col=0)
        entry = builder.create("entry", name="last_name", textvariable=tk.StringVar())
        entry.place(row=3, col=1, sticky="nsew")

        # gender
        builder.create("label", text="Gender").place(row=4, col=0)
        entry = builder.create("entry", name="gender", textvariable=tk.StringVar())
        entry.place(row=4, col=1, sticky="nsew")

        # date of birth
        builder.create("label", text="Date of Birth").place(row=5, col=0)
        entry = builder.create("entry", name="dob", textvariable=tk.StringVar())
        entry.place(row=5, col=1, sticky="nsew")

        # password
        builder.create("label", text="*Password").place(row=6, col=0)
        entry = builder.create(
            "entry", name="pw", textvariable=tk.StringVar(), show="*"
        )
        entry.place(row=6, col=1, sticky="nsew")

        # confirm pw
        builder.create("label", text="*Confirm Password").place(row=7, col=0)
        entry = builder.create(
            "entry", name="confirm_pw", textvariable=tk.StringVar(), show="*"
        )
        entry.place(row=7, col=1, sticky="nsew")

        builder.create("label", text="*required").place(row=8, col=1)
        builder.create("button2", text="Register", command=callbacks.register).place(
            row=9, col=1
        )
        return builder.view

    def clear_entries(self):
        """Clears all entries in the register view"""
        for entry in self.filter("entry"):
            entry.set_var("")

    def get_user_data(self):
        """Returns a UserSignUpData object containing data from register view entries"""
        entry_names = ["first_name", "last_name", "gender", "dob", "email"]
        user_data = user.UserSignUpData(
            *(self[f"{name}_entry"].get_var() for name in entry_names)
        )
        user_data.join_date = util.get_current_date()
        return user_data



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 17:02:53 2021

@author: richa
"""
import functools
import tkinter as tk
from typing import List

from desktop_shop import user
from desktop_shop.gui import callbacks, components


# pylint: disable=line-too-long
class View(components.View):
    """Profile view"""

    @classmethod
    def create(cls, window: components.Tk, builder: components.Builder):
        """Initialises profile view, including all its components"""
        builder.view = cls()
        builder.root = window

        frame = builder.create("frame", relief=tk.RAISED, bd=3)
        frame.place(row=2, row_span=1, col=1, col_span=3, sticky="nsew")
        builder.root = frame.component.tk_component

        # date joined
        builder.create("label", text="Date joined:").place(row=0, col=0)
        label = builder.create(
            "label", name="date_joined_data", textvariable=tk.StringVar()
        )
        label.place(row=0, col=1, sticky="w")

        # profile data frame
        builder.root = window
        frame = builder.create("frame", relief=tk.RAISED, bd=3)
        frame.place(row=3, row_span=6, col=1, col_span=3, sticky="new")
        builder.root = frame.component.tk_component

        # profile data
        _create_profile_data_entry(
            builder, text="First name:", name="first_name", row=0
        )
        _create_profile_data_entry(builder, text="Last name:", name="last_name", row=1)
        _create_profile_data_entry(builder, text="Gender:", name="gender", row=2)
        _create_profile_data_entry(builder, text="Email address:", name="email", row=3)
        _create_profile_data_entry(builder, text="Date of Birth:", name="dob", row=4)
        button = builder.create(
            "button2",
            name="edit_user_data",
            text="Confirm changes",
            command=callbacks.edit_user_data,
            bd=3,
        )
        button.place(row=5, col=1, col_span=2, sticky="nse")

        # password change
        builder.root = window
        frame = builder.create("frame", relief=tk.RAISED, bd=3)
        frame.place(row=10, col=1, col_span=3, sticky="nsew")
        builder.root = frame.component.tk_component

        button = builder.create(
            "button",
            text="Change password",
            name="password_change_frame",
            command=callbacks.show_password_change_frame,
            bd=3,
        )
        button.place(row=0, col=2, sticky="w")

        frame = builder.create("frame", name="password_change", relief=tk.RAISED, bd=0)
        frame.place(row=0, row_span=3, col=0, col_span=3)
        builder.root = frame.component.tk_component
        builder.view.hide_components("password_change_frame")

        # password
        builder.create("label", text="*Password").place(row=0, col=0)
        builder.create("entry", name="pw", textvariable=tk.StringVar(), show="*").place(
            row=0, col=1, sticky="we"
        )

        # confirm password
        builder.create("label", text="*Confirm").place(row=1, col=0)
        builder.create(
            "entry", name="confirm_pw", textvariable=tk.StringVar(), show="*"
        ).place(row=1, col=1, sticky="we")

        button = builder.create(
            "button2",
            text="Confirm password change",
            command=callbacks.edit_user_password,
            bd=3,
        )
        button.place(row=2, col=1, sticky="e")

        return builder.view

    def store_user_data(self, user_data):
        """Stores the passed user data in the user data entries"""
        for name, value in zip(self.user_data_entries, user_data):
            self[name].set_var(value)

    def get_user_data(self):
        """Returns a UserSignUpData object containing the data in the entries"""
        return user.UserSignUpData(*(self[e].get_var() for e in self.user_data_entries))

    @property
    def user_data_entries(self) -> List[str]:
        """Getter for user_data_entries, returns list of data entries in the profile view"""
        entry_names = ["first_name", "last_name", "gender", "dob", "email"]
        return [f"{name}_entry" for name in entry_names] + ["date_joined_data_label"]


def _create_profile_data_entry(
    builder: components.Builder, text: str, name: str, row: int
) -> None:
    """Creates a labelled entry + edit button"""
    builder.create("label", text=text).place(row=row, col=0, sticky="w")

    entry = builder.create("entry", name=name, textvariable=tk.StringVar())
    entry.place(row=row, col=1, sticky="we")
    entry.component.config(state="disabled")

    callback = functools.partial(callbacks.activate_profile_entry, entry.component.name)
    button = builder.create("button", text="Edit", command=callback, bd=3)
    button.place(row=row, col=2, sticky="e")



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 17:05:23 2021

@author: richa
"""
import functools

from desktop_shop import gui, server
from desktop_shop.gui import components


class View(components.View):
    """Home view"""

    @classmethod
    def create(cls, *_):
        """Initialises an empty home view. All other home view contents are generated
        in init_product_data.
        """
        return cls()

    def init_product_data(self, window: components.Tk, builder: components.Builder):
        """Dynamically generates all contents of the home view"""
        builder.view = self

        with gui.db_conn as cursor:
            product_datas = server.query_product_data_from_product_table(cursor)

        for i, (product_id, name, price) in enumerate(product_datas, 1):
            builder.root = window
            frame = builder.create("menu_frame")
            frame.place(row=i, row_span=1, col=1, col_span=3, sticky="we")
            builder.root = frame.component.tk_component

            # product name label
            label = builder.create("label", width=40, text=f"{name:>30} ${price:<7}")
            label.place(row=i, col=0, sticky="w")

            # add to cart button
            callback = functools.partial(self.add_to_cart, product_id)
            button = builder.create(
                "button", text="Add to cart", command=callback, bd=3
            )
            button.place(row=i, col=2, sticky="e")

    @staticmethod
    def add_to_cart(product_id):
        """Callback for Add to Cart button. Appends the product id to cart
        and shows checkout button
        """
        gui.app.data["cart"].append(str(product_id))
        if gui.app["main_menu"]["checkout_button"].hidden:
            gui.app["main_menu"].unhide_components("checkout_button")



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 14 16:29:44 2020

Call using parent level database.py using:

    database generate

or:
    database generate --fast

@author: Korean_Crimson
"""
import random
import sqlite3

from desktop_shop import crypto, database, util
from desktop_shop.datagen import data

# combined with every salt for extra security in pw hashing
PEPPER = "secret"

# pylint: disable=too-few-public-methods
class UserDataGenerator:
    """Used to generate data to fill the users table in main.db"""

    def __init__(self):
        self._email_cache = []

    def populate_user_table(
        self, cursor, number_of_users=10_000, hash_iterations=100_000
    ):
        """Populates the user table in main.db (needs to already exist) with the
        amount of users specified in the input arg. Expects a database connection
        object to be passed in as first arg
        """
        first_names = data.fetch_first_names()
        last_names = data.fetch_last_names()

        for _ in range(number_of_users):
            user_data = self._generate_random_user_data(first_names, last_names)
            password = self._generate_new_password()
            database.add_user(
                cursor, user_data, password, PEPPER, iterations=hash_iterations
            )

    def _generate_random_user_data(self, first_names, last_names):
        first_name = random.choice(list(first_names.keys()))
        last_name = random.choice(last_names)
        gender = first_names[first_name]
        join_date = data.get_random_date(start=[2014, 6, 1], end=[2020, 11, 1])
        dob = data.get_random_date(start=[1920, 1, 1], end=[2004, 6, 1])
        email_address = self._generate_random_email_address(first_name, last_name, dob)
        return [first_name, last_name, gender, dob, email_address, join_date]

    def _generate_random_email_address(self, first_name, last_name, dob):
        """generates unique email using _email_cache storing all generated emails"""
        email_address = ""
        domains = [
            "emailook.com",
            "oldlurk.com",
            "failmailprovider.co.uk",
            "beardgamemail.com",
        ]
        concatenators = ["-", "_", ".", "", ""]
        while not email_address or email_address in self._email_cache:
            domain = random.choice(domains)
            year, *_ = dob.split("-")
            random_string = "".join(
                [chr(random.randint(65, 90)) for _ in range(random.randint(1, 5))]
            )
            random_num = str(random.randint(0, 999))
            user = random.choices(
                [first_name, last_name, year, random_string, random_num],
                k=random.randint(2, 4),
            )
            concatenator = random.choice(concatenators)
            email_address = f"{concatenator.join(user)}@{domain}"
        self._email_cache.append(email_address)
        return email_address

    @staticmethod
    def _generate_new_password():
        return "password"


class ProductDataGenerator:
    """Used to generate data to fill the products table in main.db"""

    def populate_products_table(self, cursor, number_of_products=20):
        """Populates the products table in main.db (needs to already exist)
        with the amount of products specified in the input arg. Expects a
        database connection to be passed in as first arg
        """
        for _ in range(number_of_products):
            product_data = self._generate_random_product_data()
            database.add_product(cursor, product_data)

    @staticmethod
    def _generate_random_product_data():
        product_names = ["PumaVoxel", "PetaBit", "PersePhone"]
        series = ["A", "M", "S", "X", "Y", "Z"]
        versions = ["99", "10", "2", "33", "21"]

        name = random.choice(product_names)
        series_ = random.choice(series)
        version = random.choice(versions)
        product_name = f"{name} {series_}{version}"
        product_price = random.randint(10, 50) * 10 - 0.01  # Between 99.99 and 499.99

        product_data = [product_name, product_price]
        return product_data


class TransactionDataGenerator:
    """Used to generate data to fill the transactions table in main.db"""

    @classmethod
    def populate_transactions_table(cls, cursor, number_of_transactions=100_000):
        """Populates the transactions table in main.db (needs to already exist)
        with the amount of transactions specified in the input arg. Expects a
        database connection object to be passed in as first arg
        """
        user_ids = database.query_user_ids_from_user_table(cursor)
        product_ids = database.query_product_ids_from_product_table(cursor)

        for _ in range(number_of_transactions):
            chosen_product_ids = cls._generate_random_chosen_product_ids(product_ids)
            transaction_data = cls._generate_random_transaction_data(user_ids)
            database.add_transaction(cursor, transaction_data, chosen_product_ids)

    @staticmethod
    def _generate_random_transaction_data(user_ids):
        user_id = random.choice(user_ids)
        date = data.get_random_date(start=[2004, 6, 1], end=[2020, 11, 1])

        transaction_data = [user_id, date]
        return transaction_data

    @staticmethod
    def _generate_random_chosen_product_ids(product_ids):
        number_of_products = random.randint(1, 5)
        chosen_product_ids = random.choices(product_ids, k=number_of_products)
        return list(map(str, chosen_product_ids))


class SessionDataGenerator:
    """Used to generate data to fill the sessions table in main.db"""

    @staticmethod
    def populate_sessions_table(cursor, number_of_sessions=25):
        """Populates the sessions table in main.db (needs to already exist)
        with the amount of sessions specified in the input arg. Expects a
        database connection object to be passed in as first arg
        """
        user_ids = database.query_user_ids_from_user_table(cursor)
        chosen_user_ids = random.sample(
            user_ids, min(number_of_sessions, len(user_ids))
        )

        for user_id in chosen_user_ids:
            session_id = crypto.generate_new_session_id()
            timestamp = util.generate_timestamp()
            session_data = [session_id, user_id, timestamp]
            database.add_session(cursor, session_data)


def generate(hash_iterations=100_000, transactions=100_000, users=10_000, products=20):
    """Generates the database, including user, products, transactions,
    detailed_transactions and sessions tables.
    The tables are pre-populated with random data of the specified sizes.
    """

    with sqlite3.connect("main.db") as cursor:
        # create and populate user table
        database.create_user_table(cursor)
        UserDataGenerator().populate_user_table(
            cursor, hash_iterations=hash_iterations, number_of_users=users
        )

        # create and populate products table
        database.create_products_table(cursor)
        ProductDataGenerator().populate_products_table(
            cursor, number_of_products=products
        )

        # create detailed transactions table
        database.create_detailed_transactions_table(cursor)

        # create and populate transactions table
        database.create_transactions_table(cursor)
        TransactionDataGenerator.populate_transactions_table(
            cursor, number_of_transactions=transactions
        )

        # create and populate sessions table
        database.create_sessions_table(cursor)
        SessionDataGenerator.populate_sessions_table(cursor)


if __name__ == "__main__":
    generate()



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 10:00:48 2021

@author: Korean_Crimson
"""
import tkinter as tk
from typing import Dict

from desktop_shop import gui, user
from desktop_shop.gui import components, config
from desktop_shop.gui.views import views

# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
# pylint: disable=line-too-long


def init_root():
    """Initialises and configures tk root"""
    gui.root.title("OfflineShop")
    gui.root.wm_attributes("-transparentcolor", "purple")
    gui.root.config(bg=config.BG)
    for _ in range(21):
        gui.root.add_row(30)
    gui.root.add_col(150)
    gui.root.add_col(100)
    gui.root.add_col(100)
    gui.root.add_col(100)
    gui.root.add_col(150)


def init_builder():
    """Initialises the component builder and registers the widget factories"""
    builder = components.Builder()

    factory = components.Factory(tk.Frame, components.Frame, kwargs=config.FRAME_THEME)
    builder.register("frame", factory)

    kwargs = {"relief": tk.RAISED, "bd": 3, "bg": config.BG2}
    factory = components.Factory(tk.Frame, components.Component, kwargs=kwargs)
    builder.register("menu_frame", factory)

    factory = components.EntryFactory(
        tk.Label, components.Component, kwargs=config.LABEL_THEME
    )
    builder.register("label", factory)

    factory = components.EntryFactory(
        tk.Label, components.Component, kwargs=config.ERROR_THEME
    )
    builder.register("label2", factory)

    factory = components.EntryFactory(
        tk.Entry, components.Component, kwargs=config.ENTRY_THEME
    )
    builder.register("entry", factory)

    factory = components.Factory(
        tk.Button, components.Component, kwargs=config.BUTTON_THEME
    )
    builder.register("button", factory)

    factory = components.Factory(
        tk.Button, components.Component, kwargs=config.BUTTON_THEME2
    )
    builder.register("button2", factory)

    return builder


def init_views(
    window: components.Tk, builder: components.Builder
) -> Dict[str, components.View]:
    """Initialises all views"""
    views_: Dict[str, components.View] = {
        name: view.create(window, builder) for name, view in views.items()
    }
    views_["home"].activate()
    views_["main_menu"].activate()
    return views_


def init():
    """Init function that needs to be called before gui is started"""
    init_root()
    gui.app.builder = init_builder()
    gui.app.views_dict = init_views(gui.root, gui.app.builder)
    gui.app.data = {
        "session_id": None,
        "user_data": user.UserSignUpData(),
        "pw_hash": "",
        "cart": [],
    }
    gui.app["home"].init_product_data(gui.root, gui.app.builder)



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 17:04:48 2021

@author: richa
"""
import tkinter as tk

from desktop_shop.gui import callbacks, components


class View(components.View):
    """Login view"""

    @classmethod
    def create(cls, window: components.Tk, builder: components.Builder):
        """Initialises login view, including all its components"""
        builder.view = cls()
        builder.root = window

        # login frame
        frame = builder.create("frame", relief=tk.SUNKEN, bd=2)
        frame.place(row=8, row_span=4, col=1, col_span=3, sticky="nsew")
        builder.root = frame.component.tk_component

        builder.create("label", text="Login").place(row=0, col=1)

        # email
        builder.create("label", text="Email address").place(row=1, col=0)
        edit = builder.create("entry", name="email", textvariable=tk.StringVar())
        edit.place(row=1, col=1, sticky="nsew")

        # password
        builder.create("label", text="Password").place(row=2, col=0)
        edit = builder.create("entry", name="pw", textvariable=tk.StringVar(), show="*")
        edit.place(row=2, col=1, sticky="nsew")

        builder.create("button2", text="Log in", command=callbacks.login).place(
            row=3, col=1
        )

        # login unsuccessful
        message = "Logging in has failed. Please check your credentials."
        label = builder.create("label2", name="login_failed", text=message)
        label.place(row=4, col=0, col_span=3)
        builder.view.hide_components("login_failed_label")

        return builder.view

    def clear_entries(self):
        """Clears the password and email entries"""
        for entry in ["pw_entry", "email_entry"]:
            self[entry].set_var("")



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 10:01:35 2021

@author: Korean_Crimson
"""

TITLE = "Offline Shop"

# colours
FG = "#FFFFFF"
DISABLED_FG = "#EEEEEE"
BG = "#121212"
BG2 = "#242424"
BG3 = "#363636"
BG4 = "#484848"
BG5 = "#606060"
PRIM = "#3700B3"
SEC = "#A172E1"
ERR = "#CF6679"

# themes
ENTRY_THEME = {
    "fg": FG,
    "bg": BG4,
    "disabledbackground": BG5,
    "disabledforeground": DISABLED_FG,
}
FRAME_THEME = {"bg": BG2, "highlightbackground": BG3}
THEME = {**FRAME_THEME, "fg": FG}

LABEL_THEME = {"fg": FG, "bg": BG2, "highlightbackground": BG2}
BUTTON_THEME = {"fg": FG, "bg": BG3, "activebackground": BG4, "activeforeground": FG}
BUTTON_THEME2 = {
    "fg": FG,
    "bg": PRIM,
    "activebackground": BG4,
    "activeforeground": FG,
    "disabledforeground": DISABLED_FG,
}
DROPDOWN_THEME = {"fg": FG, "bg": BG2, "activebackground": BG2, "activeforeground": FG}
ERROR_THEME = {"fg": ERR, "bg": BG2}



# -*- coding: utf-8 -*-
"""
Created on Mon Feb  1 09:53:11 2021

@author: Korean_Crimson
"""
from __future__ import annotations

import re
import tkinter as tk
import uuid
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union

Root = Union[tk.Tk, tk.Widget]


class Tk(tk.Tk):
    """Wrapper around tk.Tk class with easier rowconfigure and columnconfigure functionality"""

    def __init__(self):
        super().__init__()
        self.col_num: int = 0
        self.row_num: int = 0

    def add_row(self, minsize: int, weight: int = 1) -> None:
        """Wraps around tk.Tk.rowconfigure"""
        self.rowconfigure(self.row_num, minsize=minsize, weight=weight)
        self.row_num += 1

    def add_col(self, minsize: int, weight: int = 1) -> None:
        """Wraps around tk.Tk.columnconfigure"""
        self.columnconfigure(self.col_num, minsize=minsize, weight=weight)
        self.col_num += 1


@dataclass
class Gui:
    """Class to contain all tk widgets, tk Vars and other data related to the gui"""

    window: Tk
    views_dict: Dict[str, View] = field(default_factory=dict)
    data: Dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        self.builder: Optional[Builder] = None

    def __enter__(self):
        self.pack_all()
        return self

    def __exit__(self, *_):
        pass

    def __getitem__(self, key: str) -> Optional[View]:
        return self.views_dict.get(key)

    def mainloop(self) -> None:
        """Wraps around tk.Tk.mainloop"""
        self.window.mainloop()

    def pack_all(self) -> None:
        """Places all active views on the gui by calling pack"""
        for view in self.views_dict.values():
            if view.active:
                view.pack()
            else:
                view.unpack()

    def deactivate_all(self) -> None:
        """Deactivates all views"""
        for view in self.views_dict.values():
            view.deactivate()

    def switch_to(self, *keys: str) -> None:
        """Switches to a particular view (deactivates all other views)"""
        self.deactivate_all()
        for key in keys:
            self.views_dict[key].activate()
        self.pack_all()


class View:
    """View class that acts as a container of a number of related Component objects"""

    def __init__(self):
        self.active = False
        self._components: Dict[str, Component] = {}
        self._entries = {}
        self._frame_components: Dict[str, Component] = {}

    def __getitem__(self, component_name):
        """Returns the component identified by the passed name"""
        return self._all_components[component_name]

    def get_frames(self) -> Iterable[Component]:
        """Returns all Components that are marked as frames"""
        return self._frame_components.values()

    def clear(self) -> None:
        """Clears the current view by unpacking all components and clearing all
        component dicts (uninitialises the entire view).
        """
        self.unpack()
        self._components = {}
        self._frame_components = {}

    def filter(self, value: str) -> List[Component]:
        """Returns components containing the passed string in their name"""
        return [
            c
            for c in self._all_components.values()
            if c.name is not None and value in c.name
        ]

    def activate(self) -> None:
        """Activates the current view (when the Gui object next calls pack on it,
        this View will show when active)
        """
        self.active = True

    def deactivate(self) -> None:
        """Deactivates the current view (when the Gui object next calls pack on it,
        this View will not show when deactive)
        """
        self.active = False

    def add_component(self, component: Component, name: str) -> None:
        """Adds the passed Component under the specified name and stores it in the view"""
        self._components[name] = component

    def add_entry_component(self, component: Component, name: str) -> None:
        """Adds the passed entry component under the specified name and stores it in the view"""
        self._components[f"{name}_entry"] = component
        self._entries[name] = component

    def add_frame_component(self, component: Component, name: str) -> None:
        """Adds the passed Component under the specified name and stores it in the view, as
        a component of type frame
        """
        self._frame_components[name] = component

    def pack(self) -> None:
        """Calls gridpack on all components in the view (frames first)"""
        for component in self._all_components.values():
            component.gridpack()

    def repack(self) -> None:
        """Unpacks, then repacks the view"""
        self.unpack()
        self.pack()

    def unpack(self) -> None:
        """Calls unpack for all components in the view (frames first)"""
        for component in self._all_components.values():
            component.unpack()

    def hide_components(self, *component_names: str) -> None:
        """Hides a particular component"""
        for component_name in component_names:
            self._all_components[component_name].hide()

    def unhide_components(self, *component_names: str) -> None:
        """Unhides a particular component"""
        for component_name in component_names:
            component = self._all_components[component_name]
            component.unhide()
            component.gridpack()

    @property
    def _all_components(self) -> Dict[str, Component]:
        return {**self._components, **self._frame_components}


# pylint: disable=too-many-instance-attributes
@dataclass
class Component:
    """Wrapper class around tk widgets, that stores all data usually required when
    calling .grid(). Using this class makes it possible to repeatedly call grid()
    and grid_forget() for variable gui layouts without passing the arguments repeatedly
    """

    tk_component: tk.Widget
    row: int = 0
    col: int = 0
    sticky: str = "n"
    padx: int = 0
    pady: int = 0
    col_span: int = 1
    row_span: int = 1
    var: Any = None
    name: str = ""

    def __post_init__(self):
        self.data = None
        self.hidden = False

    def hide(self) -> None:
        """Hides the component and calls unpack"""
        self.hidden = True
        self.unpack()

    def unhide(self) -> None:
        """Unhides the component"""
        self.hidden = False

    def add_to(self, view: View) -> None:
        """Adds the component to the passed view"""
        view.add_component(self, self.name)

    def gridpack(self) -> None:
        """calls .grid on the tk_component if it is not hidden"""
        if self.hidden:
            return

        self.tk_component.grid(
            row=self.row,
            column=self.col,
            sticky=self.sticky,
            padx=self.padx,
            pady=self.padx,
            rowspan=self.row_span,
            columnspan=self.col_span,
        )

    def unpack(self) -> None:
        """calls .grid_forget on the tk_component"""
        self.tk_component.grid_forget()

    def get(self) -> Any:
        """calls .get on the tk_component"""
        return self.tk_component.get()  # type: ignore # pylint: disable=no-member

    def get_var(self) -> Any:
        """calls .get on the stored tk Var"""
        return self.var.get() if self.var else None  # type: ignore

    def set_var(self, value) -> None:
        """calls .set on the stored tk Var"""
        if self.var:
            self.var.set(value)  # type: ignore

    def config(self, *args, **kwargs) -> None:
        """calls .config on the tk_component"""
        self.tk_component.config(*args, **kwargs)  # type: ignore


class Frame(Component):
    """Frame compoenent"""

    def add_to(self, view: View) -> None:
        """Adds the frame component to the view. Overrides Component.add_to"""
        view.add_frame_component(self, self.name)


@dataclass
class ComponentPlacer:
    """Component placer class, returned from Factory.create"""

    component: Component

    def place(self, **kwargs: Any):
        """Places the component using the specified keyword arguments"""
        for key, value in kwargs.items():
            setattr(self.component, key, value)


@dataclass
class Factory:
    """Factory object, contains a method and a respective component class,
    as well as creation arguments and keyword arguments.
    """

    method: Callable[..., tk.Widget]
    component: Type
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def create(self, root: Root, name: str = None, **kwargs: Any) -> ComponentPlacer:
        """Creates a new Component by calling the creation method. Returns a ComponentPlacer"""
        widget = self.method(root, *self.args, **self.kwargs, **kwargs)
        component = self.component(widget, name=name)
        return ComponentPlacer(component)


# pylint: disable=too-few-public-methods
class EntryFactory(Factory):
    """Entry factory"""

    def create(self, root: Root, name: str = None, **kwargs: Any) -> ComponentPlacer:
        """Overrides Factory.create.
        Adds an extra textvariable to the widget creation process.
        """
        component_placer = super().create(root, name, **kwargs)
        component_placer.component.var = kwargs.get("textvariable")
        return component_placer


@dataclass
class Builder:
    """Component builder class"""

    factory_methods: Dict[str, Factory] = field(default_factory=dict)

    def __post_init__(self):
        self.view: View = None
        self.root: Root = None

    def register(self, method_name: str, factory: Factory) -> None:
        """Register a factory under a specified name"""
        self.factory_methods[method_name] = factory

    def create(
        self, method_name: str, name: str = None, **kwargs: Any
    ) -> ComponentPlacer:
        """Creates a component (widget) by looking up the passed method_name
        from the registered factories and calling create on it.
        Passes received keyword arguments to the factory create method.
        Returns the component's ComponentPlacer.
        """
        factory = self.factory_methods[method_name]

        method_name_ = re.sub("[0-9]+", "", method_name)
        name = f"{name}_{method_name_}" if name is not None else str(uuid.uuid4())

        component_placer = factory.create(self.root, name, **kwargs)
        component_placer.component.add_to(self.view)
        return component_placer



# -*- coding: utf-8 -*-
"""Setup file for pip install"""
from pathlib import Path

import setuptools

project_dir = Path(__file__).parent

setuptools.setup(
    name="desktop_shop",
    version="1.0.3",
    description="Desktop shop application",
    long_description=project_dir.joinpath("README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    keywords=["python"],
    author="Richard Baltrusch",
    url="https://github.com/rbaltrusch/desktop_shop",
    packages=setuptools.find_packages("."),
    package_dir={"": "."},
    python_requires=">=3.8",
    include_package_data=True,
    package_data={"desktop_shop": ["py.typed"]},  # for mypy
    # This is a trick to avoid duplicating dependencies between both setup.py and requirements.txt.
    # requirements.txt must be included in MANIFEST.in for this to work.
    install_requires=project_dir.joinpath("requirements.txt")
    .read_text(encoding="utf-8")
    .split("\n"),
    zip_safe=False,
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Typing :: Typed",
    ],
)



# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 16:57:41 2021

@author: richa
"""
import tkinter as tk

from desktop_shop.gui import callbacks, components, config


# pylint: disable=line-too-long
class View(components.View):
    """Main menu view"""

    @classmethod
    def create(cls, window: components.Tk, builder: components.Builder):
        """Initialises main menu view, including all its components"""
        builder.view = cls()
        builder.root = window

        frame = builder.create("menu_frame")
        frame.place(row=0, col=0, col_span=5, row_span=1, sticky="nsew")
        builder.root = frame.component.tk_component

        # buttons
        builder.create(
            "button", name="home", command=callbacks.switch_to_home, text="Home"
        ).place(row=0, col=0)
        builder.create(
            "button",
            name="register",
            command=callbacks.switch_to_register,
            text="Register",
        ).place(row=0, col=1)
        builder.create(
            "button", name="login", command=callbacks.switch_to_login, text="Login"
        ).place(row=0, col=2, sticky="nsew")

        button = builder.create(
            "button",
            name="checkout",
            text="Checkout",
            command=callbacks.switch_to_checkout,
        )
        button.place(row=0, col=3)
        builder.view.hide_components("checkout_button")

        # logged in as frame
        logged_in_as_frame = tk.Frame(
            window, borderwidth=0, highlightthickness=0, bg=config.BG2
        )
        component = components.Component(logged_in_as_frame, row=0, col=4, sticky="e")
        builder.view.add_frame_component(component, "logged_in_as_frame")
        builder.view.hide_components("logged_in_as_frame")

        # options dropdown
        variable = tk.StringVar()
        options = ["Profile", "Sign out"]
        variable.set("Logged in as {}")
        variable.trace("w", callbacks.on_dropdown_value_write_event)
        options_dropdown = tk.OptionMenu(logged_in_as_frame, variable, *options)
        options_dropdown["menu"].config(**config.DROPDOWN_THEME)
        component = components.Component(
            options_dropdown, row=0, col=7, col_span=2, sticky="nsew", var=variable
        )
        component.config(**config.DROPDOWN_THEME)
        component.config(highlightthickness=0, bd=3)
        builder.view.add_component(component, "options_dropdown")

        # message frame
        builder.root = window
        frame = builder.create("frame", name="message", relief=tk.RIDGE, bd=0)
        frame.place(row=0, col=3, col_span=1, sticky="w")
        builder.view.hide_components("message_frame")
        builder.root = frame.component.tk_component

        builder.create("label", name="message", textvariable=tk.StringVar()).place(
            row=0, col=0
        )
        builder.create(
            "label2", name="error_message", textvariable=tk.StringVar()
        ).place(row=0, col=0)

        return builder.view
